<?php

namespace Drupal\entrypoints\Plugin;

use Drupal\Component\Plugin\PluginBase;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Process\Process;

/**
 * Base class for entrypoint runtime plugins.
 */
abstract class EntrypointsRuntimePluginBase extends PluginBase implements EntrypointsRuntimePluginInterface, ContainerFactoryPluginInterface {

  /**
   * The alias or absolute path to the executable binary.
   *
   * @var string
   */
  protected $bin;

  /**
   * Internal in-memory cache.
   *
   * @var array
   */
  protected $cache;

  /**
   * The logger channel factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * The JSON serialization service, only available at the service instance.
   *
   * @var \Drupal\Component\Serialization\SerializationInterface
   */
  protected $jsonSerialization;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = new static($configuration, $plugin_id, $plugin_definition);

    $instance->reset();
    $instance->loggerFactory = $container->get('logger.factory');
    $instance->jsonSerialization = $container->get('serialization.json');
    $instance->bin = $plugin_id;
    $bin_path = $container->getParameter('entrypoints.bin_path');
    if (!empty($bin_path)) {
      if (substr($bin_path, -strlen(DIRECTORY_SEPARATOR)) === DIRECTORY_SEPARATOR) {
        $instance->bin = $bin_path . $plugin_id;
      }
      else {
        $instance->bin = $bin_path . DIRECTORY_SEPARATOR . $plugin_id;
      }
    }

    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getVersion() {
    $process = new Process([$this->bin, '-v']);
    $process->run();
    if (!$process->isSuccessful()) {
      return NULL;
    }
    if ($output = strtok($process->getIncrementalOutput(), "\t\n")) {
      $output = trim($output);
    }
    if (empty($output)) {
      return NULL;
    }

    return preg_match('/^\d+\.\d+\.\d+$/', $output) === 1 ? $output : NULL;
  }

  /**
   * Builds an executable process object.
   *
   * @param array $arguments
   *   The arguments for the runtime.
   * @param string|null $cwd
   *   The working directory or null to use the working dir of the current PHP
   *   process.
   * @param array|null $env
   *   The environment variables or null to use the same environment as the
   *   current PHP process.
   * @param mixed|null $input
   *   The input as stream resource, scalar or \Traversable, or null for no
   *   input.
   * @param int|float|null $timeout
   *   (Optional) The timeout in seconds, 0 or NULL for unlimited time. The
   *   limitation factor might be cut to the current maximum execution time.
   *
   * @return \Symfony\Component\Process\Process
   *   The wrapped process object, ready to be executed.
   */
  protected function buildExecutableProcess(array $arguments, $cwd = NULL, array $env = NULL, $input = NULL, $timeout = NULL) {
    if ($timeout == 0) {
      $timeout = NULL;
    }
    if ($max_execution_time = $this->getMaxExecutionTime()) {
      if ($max_execution_time > 0 && ($timeout === NULL || $max_execution_time < $timeout)) {
        if ($max_execution_time < 240 && empty($this->cache['warning_emitted'])) {
          $this->logger()->warning("A very limited time window for the entrypoints runtime execution was detected. Please avoid a too much limited max_execution_time resource when using the entrypoints runtime. Make sure that long-running entrypoint runtime executions will not be run on regular public site requests.");
          $this->cache['warning_emitted'] = TRUE;
        }
        $timeout = $max_execution_time > 25 ? $max_execution_time - 5 : $max_execution_time - 1;
        if ($timeout <= 0 && empty($this->cache['warning_emitted'])) {
          $this->logger()->critical("Too limited time window for the entrypoints runtime execution detected. The execution time is most likely too low. Please make sure to have a sufficient max_execution time available.");
          $this->cache['warning_emitted'] = TRUE;
          $timeout = 0.5;
        }
      }
    }
    $command = array_merge([$this->bin], $arguments);
    return new Process($command, $cwd, $env, $input, $timeout);
  }

  /**
   * Get a logger channel.
   *
   * @param string $channel
   *   The channel to get. Default set to the 'entrypoints' channel.
   *
   * @return \Drupal\Core\Logger\LoggerChannelInterface
   *   The required logger channel.
   */
  protected function logger($channel = 'entrypoints') {
    return $this->loggerFactory->get($channel);
  }

  /**
   * Get the maximum execution time setting value.
   *
   * @return int
   *   The maximum execution time setting value.
   */
  protected function getMaxExecutionTime() {
    if (!isset($this->cache['max_execution_time'])) {
      $this->cache['max_execution_time'] = (int) ini_get('max_execution_time');
    }
    return $this->cache['max_execution_time'];
  }

  /**
   * {@inheritdoc}
   */
  public function reset() {
    $this->cache = [];
  }

}
